{{

  Wolf Spyder flight control code

}}  

CON

  _xinfreq = 5_000_000                      
  _clkmode = xtal1 | pll16x
  PID_ROLL  = 0
  PID_PITCH = 1

  IN_THRO = 0
  IN_AILE = 1
  IN_ELEV = 2
  IN_RUDD = 3
  IN_GEAR = 4
  IN_FLAP = 5


  OUT_ESC_FR = 0                ' Front Right ESC                       
  OUT_ESC_FL = 1                ' Front Left  ESC
  OUT_ESC_BR = 2                ' Back  Right ESC                        
  OUT_ESC_BL = 3                ' Back  Left  ESC

  OUT_SRV_MR = 4                ' Right Motor Servos
  OUT_SRV_ML = 5                ' Left  Motor Servos
  
  OUT_SRV_RA = 6                ' Right Aileron                       
  OUT_SRV_LA = 7                ' Left  Aileron
  OUT_SRV_RE = 8                ' Right Elevator
  OUT_SRV_LE = 9                ' Left  Elevator

  OUT_SRV_RUDD = 10             ' Rudder servo                        
  OUT_SRV_GEAR = 11             ' Gear servo                          


  OPT_NUM_OPTIONS = 4
  OPT_PID_PGAIN = 0
  OPT_PID_IGAIN = 1
  OPT_RUDD_GAIN = 2
  OPT_RC_IN_GAIN = 3


VAR

  long Thro, Aile, Elev, Rudd, Gear, Flap               ' RC receiver inputs as locals
  long AngleX, AngleY

  long FlightMix                                        ' 0 - 256 value for mixing between regular flight and stabilized hover
  long Output[12]                                       ' Temp output flight controls array, copied into 'servo' when finished                        
  

OBJ

  s             : "Servo32v3Mod.spin"
  RC            : "RC_Receiver_6.spin"
  Kalman        : "KalmanManager.spin" 
  pid[2]        : "PIDFast"
  Serial        : "Simple_Serial.spin"
  Num           : "Simple_Numbers.spin"


PUB init

  ' Initialize servo values to all neutral (zero for throttles)
  ' Delay values of 1500 = center, 1000 = min, 2000 = max
  Output[0] := 1100    
  Output[1] := 1100              ' motors off    
  Output[2] := 1100    
  Output[3] := 1100
      
  Output[4] := 1500    
  Output[5] := 1500    
  Output[6] := 1500              ' servos neutral    
  Output[7] := 1500    
  Output[8] := 1500    
  Output[9] := 1500    
  Output[10] := 1500    
  Output[11] := 1000             ' gear down    


  'Start the servo driver cog
  s.AddPin(8)
  s.AddPin(9)
  s.AddPin(10)
  s.AddPin(11)  
  s.AddPin(12)
  s.AddPin(13)
  s.AddPin(14)
  s.AddPin(15)  
  s.Set(  8, Output[0] )
  s.Set(  9, Output[1] )
  s.Set( 10, Output[2] )
  s.Set( 11, Output[3] )
  s.Set( 12, Output[4] )
  s.Set( 13, Output[5] )
  s.Set( 14, Output[6] )
  s.Set( 15, Output[7] )
  s.Start

  'Initialize the RC input object
  RC.setpins( %111111 )
  RC.start  

  Serial.init(31, 30, 9600)


  'Initialize PID object.  Values were determined through experimentation.
  '                                               integral
  '                                      set      clamps
  '                     kp     ki        point    high low
  pid[PID_ROLL ].init( 2.1,    12.02,    0.0,     1000.0, -1000.0)
  pid[PID_PITCH].init( 2.1,    12.02,    0.0,     1000.0, -1000.0)


  'Call the main method (otherwise, the program would stop here).

  'Wait until we get a valid RC input value
  Elev := 0
  repeat while ( RC.get(IN_THRO) == 0 )
    Serial.tx(".")
    if( ++Elev == 16 )
      Serial.tx( 254 )
      Serial.tx( 1 )                                    ' clear
      Elev := 0
    waitcnt( cnt + 8000000 )                            ' wait 1/10th sec

  Kalman.Start  

  Main                                      



PUB Main | dist[2], Counter, ThroMix, Kx, Ky, Gx, Gy, Ax, Ay, Acx, Acy, DesiredX, DesiredY, RudderGain, ControlGain, ProgCount

  {
  repeat
    'Get distances
    dist[L] := ir[L].Distance 
    dist[R] := ir[R].Distance 

    'Run distances through PID objects.
    'The outputs are used to update the servo drive variables, which are constantly monitored
    'by the Servos object.  As soon as the variables change, the servos object updates the pulse
    'train to the servos.
    servo[L] := pid[L].calculate(dist[L])
    servo[R] := pid[R].calculate(dist[R])
  }

  pid[0].updateFromInt(0, Options[OPT_PID_PGAIN])  
  pid[0].updateFromInt(1, Options[OPT_PID_IGAIN])  
  
  pid[1].updateFromInt(0, Options[OPT_PID_PGAIN])  
  pid[1].updateFromInt(1, Options[OPT_PID_IGAIN])  

  RudderGain := Options[OPT_RUDD_GAIN]
  ControlGain := Options[OPT_RC_IN_GAIN] 


  Counter := 0
  ProgCount := 0
  
  repeat
    'Get the RC input values
    Thro := RC.getrc( IN_THRO )
    Aile := RC.getrc( IN_AILE )
    Elev := RC.getrc( IN_ELEV )
    Rudd := RC.getrc( IN_RUDD )
    Gear := RC.getrc( IN_GEAR )
    Flap := RC.getrc( IN_FLAP )

    'Compute the amount used to blend the two flight modes together 0 = fly, 256 = hover
    'if( Flap < -200 )
    '  FlightMix := 0
    'elseif( Flap > 200 )
      FlightMix := 256
    'else
    '  FlightMix := ((Flap + 200) * 256) / 400

    if( FlightMix == 256 AND Thro < -325 AND Rudd < -325 AND Elev > 325 AND Aile < -325 )
      ProgCount++
      if( ProgCount > 300 )
        Program

        pid[0].updateFromInt(0, Options[OPT_PID_PGAIN])  
        pid[0].updateFromInt(1, Options[OPT_PID_IGAIN])  
         
        pid[1].updateFromInt(0, Options[OPT_PID_PGAIN])  
        pid[1].updateFromInt(1, Options[OPT_PID_IGAIN])  
         
        RudderGain := Options[OPT_RUDD_GAIN]
        ControlGain := Options[OPT_RC_IN_GAIN] 
              
        next
    else
      ProgCount := 0


    'Basic flight controls (performed no matter what - stabilization added later)
    Output[OUT_ESC_FR] := Thro    
    Output[OUT_ESC_FL] := Thro     
    Output[OUT_ESC_BR] := Thro    
    Output[OUT_ESC_BL] := Thro    

    Output[OUT_SRV_ML] := Flap
    Output[OUT_SRV_MR] := Flap

     
    'Scale Aile/Elev/Rudd values for sensitivity
    Output[OUT_SRV_LA] := Aile
    Output[OUT_SRV_RA] := -Aile                       
    Output[OUT_SRV_LE] := Elev
    Output[OUT_SRV_RE] := Elev
     
    Output[OUT_SRV_RUDD] := Rudd                        
    Output[OUT_SRV_GEAR] := Gear                          


    if( FlightMix <> 0 )      

      ThroMix := (Thro + 500) >> 2                      ' Approx 0 - 256
      ThroMix <#= 128                                   ' Above 1/2 throttle, clamp it to 128
      ThroMix #>= 0

      Kalman.Update

      DesiredX := Aile * ControlGain * 4                ' These will be scaled down by PID.UpdateFromInt
      DesiredY := Elev * ControlGain * 4

      Kx := Kalman.GetX 
      Ky := -Kalman.GetY 

      Gx := Kalman.GetGyroX                        
      Gy := -Kalman.GetGyroY

      Acx := Kalman.GetAccelX
      Acy := -Kalman.GetAccelY

      
      ' Update PIDs    
      ' PID 'measured' values will be a mix of current set point value and real measurement, based on FlightMix value
      ' Set Point will be thetaX & thetaY from controller indicating desired inclination 

      pid[0].UpdateFromInt(2, DesiredX) 
      pid[1].UpdateFromInt(2, DesiredY) 
      
      AngleX := pid[0].calculate(Gx)
      AngleY := pid[1].calculate(Gy)

      ' For debugging the Kalman filter output ($8080 is a 'reset' header for my C# app)
'      if( (Counter & 3) == 0 )
'        Serial.tx( 128 )      
'        Serial.tx( 128 )      
'        Serial.tx( kx >> 8 )
'        Serial.tx( kx & 255 )
'        Serial.tx( gx >> 8 )
'        Serial.tx( gx & 255 )
'        Serial.tx( acx >> 8 )
'        Serial.tx( acx & 255 )
        
      Counter++

      ' Mix RC inputs / PID values into motor outputs using FlightMix amount
      ' FlightMix amount is currently fixed, and therefore ignored.  Once the
      ' rest of the flight control code goes in it'll be used again.

      Rudd := (Rudd * RudderGain) / 2048

      ' Since we're using Torque Yaw, feed rudder value into diagonally opposite motors
      ' else feed rudder into left/right motor servos
       
      Output[OUT_ESC_FR] += ((AngleX - AngleY - Rudd) * ThroMix) / 128  
      Output[OUT_ESC_FL] += ((-AngleX - AngleY + Rudd) * ThroMix) / 128  
      Output[OUT_ESC_BR] += ((AngleX + AngleY + Rudd) * ThroMix) / 128
      Output[OUT_ESC_BL] += ((-AngleX + AngleY - Rudd) * ThroMix) / 128


    'add 1500 to all Output values to make them 'servo friendly' again
    Output[0] += 1500 
    Output[1] += 1500 
    Output[2] += 1500 
    Output[3] += 1500 
    'Output[4] += 1500                                  ' These servos are currently not set, or used
    'Output[5] += 1500 
    'Output[6] += 1500 
    'Output[7] += 1500 
    'Output[8] += 1500 
    'Output[9] += 1500 
    'Output[10] += 1500
    'Output[11] += 1500

    Output[0] #>= 1000
    Output[0] <#= 2000
    Output[1] #>= 1000
    Output[1] <#= 2000
    Output[2] #>= 1000
    Output[2] <#= 2000
    Output[3] #>= 1000
    Output[3] <#= 2000
     
    'Copy new Ouput array into servo values
    'LONGMOVE( @servo, @Output, 12 )
    s.Set(  8, Output[0] )
    s.Set(  9, Output[1] )
    s.Set( 10, Output[2] )
    s.Set( 11, Output[3] )



PUB Program | OptionIndex, LastOption, OptionValue, OptRange

'' The programming function - Hold the RC throttle stick to lower-right
'' and the aileron/elev stick to the upper right for a few seconds to enter
'' programming mode.  Values are not persisted across power-down, and the
'' display is a 3-wire serial LCD from SparkFun.
'' While in program mode, throttle up/down selects the option, and aileron
'' left/right changes the value.  Hold both throttle and elevator full down
'' to exit.  Note that while IN program mode, output values to the motor controls
'' are kept at zero for safety.

  OptRange := 800 / (OPT_NUM_OPTIONS+1)
  LastOption := -1

  're-init the LCD    
  Serial.tx( 124 )
  Serial.tx( 132 )                                  ' very low backlight (128 + 0 to 29)
  waitcnt( cnt + 200000 ) 
   
  OptionStrings[0] := @OStr_PI_PGain 
  OptionStrings[1] := @OStr_PI_IGain 
  OptionStrings[2] := @OStr_Rudd_Gain  
  OptionStrings[3] := @OStr_RC_Gain   
  OptionStrings[4] := @OStr_Exit      
  
  'Use thro up/down to choose an option, aile right/left to inc/dec it

  repeat
    Thro := 400 - RC.getrc( IN_THRO )
    Aile := -RC.getrc( IN_AILE )
    Elev := RC.getrc( IN_ELEV )
    
    OptionIndex := Thro / OptRange
    OptionIndex #>= 0
    OptionIndex <#= OPT_NUM_OPTIONS
    OptionValue := Options[ OptionIndex ]

    if( OptionIndex <> LastOption )
      Serial.tx( 254 )
      Serial.tx( 1 )                                    ' clear
      waitcnt( cnt + 200000 ) 
       
      Serial.tx( 254 )
      Serial.tx( 128 + 0 )                              ' Cursor pos 0, line 0
      waitcnt( cnt + 200000 )
       
      Serial.str( OptionStrings[OptionIndex] )
      LastOption := OptionIndex

    OptionValue += Aile / 32
    OptionValue <#= 4096     
    OptionValue #>= 0     

    if( OptionIndex < OPT_NUM_OPTIONS )      
      Options[ OptionIndex ] := OptionValue

      Serial.tx( 254 )
      Serial.tx( 128 + 64 )                             ' Cursor pos 0, line 1
      waitcnt( cnt + 200000 )
       
      Serial.str( num.decf( OptionValue , 4 ) )
    elseif( Elev < -200 )
      quit                                              ' Stop repeating

    waitcnt( cnt + 15000000 )    ' Approx 6fps

  'Clear the LCD
  Serial.tx( 254 )
  Serial.tx( 1 )                                        ' clear

  'Write the results to the EEPROM
  
  return


DAT
Options                 long    36, 8, 512, 1024                             ' Values are typically in 2048ths

OStr_PI_PGain           byte    "P Gain",0
OStr_PI_IGain           byte    "I Gain",0
OStr_Rudd_Gain          byte    "Rudder Gain",0
OStr_RC_Gain            byte    "RC Input Gain",0
OStr_Exit               byte    "Exit (Elev Down)",0

OptionStrings           long    0, 0, 0, 0, 0